"""Parser integration tests."""

import os
import unittest
import datetime as dt

from trollsift.parser import Parser


class TestParser(unittest.TestCase):
    def setUp(self):
        self.fmt = "/somedir/{directory}/hrpt_{platform:4s}{platnum:2s}" + "_{time:%Y%m%d_%H%M}_{orbit:05d}.l1b"
        self.string = "/somedir/otherdir/hrpt_noaa16_20140210_1004_69022.l1b"
        self.data = {
            "directory": "otherdir",
            "platform": "noaa",
            "platnum": "16",
            "time": dt.datetime(2014, 2, 10, 10, 4),
            "orbit": 69022,
        }
        self.p = Parser(self.fmt)

    def test_parse(self):
        # Run
        result = self.p.parse(self.string)
        # Assert
        self.assertDictEqual(result, self.data)

    def test_cache_clear(self):
        """Test we can clear the internal cache properly"""
        from trollsift.parser import purge, regex_format

        # Run
        result = self.p.parse(self.string)
        # Assert
        self.assertDictEqual(result, self.data)
        assert regex_format.cache_info()[-1] != 0
        purge()
        assert regex_format.cache_info()[-1] == 0

    def test_compose(self):
        # Run
        result = self.p.compose(self.data)
        # Assert
        self.assertEqual(result, self.string)

    def test_validate(self):
        # These cases are True
        self.assertTrue(self.p.validate("/somedir/avhrr/2014/hrpt_noaa19_20140212_1412_12345.l1b"))
        # These cases are False
        self.assertFalse(self.p.validate("/somedir/bla/bla/hrpt_noaa19_20140212__1412_00000.l1b"))

    def assertDictEqual(self, a, b):
        for key in a:
            self.assertTrue(key in b)
            self.assertEqual(a[key], b[key])

        self.assertEqual(len(a), len(b))

    def assertItemsEqual(self, a, b):
        for i in range(len(a)):
            if isinstance(a[i], dict):
                self.assertDictEqual(a[i], b[i])
            else:
                self.assertEqual(a[i], b[i])
        self.assertEqual(len(a), len(b))


class TestParserVariousFormats(unittest.TestCase):
    def test_parse_viirs_sdr(self):
        fmt = (
            "SVI01_{platform_shortname}_d{start_time:%Y%m%d_t%H%M%S%f}_"
            "e{end_time:%H%M%S%f}_b{orbit:5d}_c{creation_time:%Y%m%d%H%M%S%f}_{source}.h5"
        )
        filename = "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"
        data = {
            "platform_shortname": "npp",
            "start_time": dt.datetime(2012, 2, 25, 18, 1, 24, 500000),
            "orbit": 1708,
            "end_time": dt.datetime(1900, 1, 1, 18, 2, 48, 700000),
            "source": "noaa_ops",
            "creation_time": dt.datetime(2012, 2, 26, 0, 21, 30, 255476),
        }
        p = Parser(fmt)
        result = p.parse(filename)
        self.assertDictEqual(result, data)

    def test_parse_iasi_l2(self):
        fmt = (
            "W_XX-EUMETSAT-{reception_location},{instrument},{long_platform_id}+{processing_location}_"
            "C_EUMS_{processing_time:%Y%m%d%H%M%S}_IASI_PW3_02_{platform_id}_{start_time:%Y%m%d-%H%M%S}Z_"
            "{end_time:%Y%m%d.%H%M%S}Z.hdf"
        )
        filename = (
            "W_XX-EUMETSAT-kan,iasi,metopb+kan_C_EUMS_20170920103559_IASI_PW3_02_"
            "M01_20170920-102217Z_20170920.102912Z.hdf"
        )
        data = {
            "reception_location": "kan",
            "instrument": "iasi",
            "long_platform_id": "metopb",
            "processing_location": "kan",
            "processing_time": dt.datetime(2017, 9, 20, 10, 35, 59),
            "platform_id": "M01",
            "start_time": dt.datetime(2017, 9, 20, 10, 22, 17),
            "end_time": dt.datetime(2017, 9, 20, 10, 29, 12),
        }
        p = Parser(fmt)
        result = p.parse(filename)
        self.assertDictEqual(result, data)

    def test_parse_olci_l1b(self):
        fmt = os.path.join(
            "{mission_id:3s}_OL_1_{datatype_id:_<6s}_{start_time:%Y%m%dT%H%M%S}_"
            "{end_time:%Y%m%dT%H%M%S}_{creation_time:%Y%m%dT%H%M%S}_{duration:4d}_"
            "{cycle:3d}_{relative_orbit:3d}_{frame:4d}_{centre:3s}_{platform_mode:1s}_"
            "{timeliness:2s}_{collection:3s}.SEN3",
            "{dataset_name}_radiance.nc",
        )
        # made up:
        filename = os.path.join(
            "S3A_OL_1_EFR____20180916T090539_20180916T090839_20180916T090539_0001_001_001_0001_CEN_M_AA_AAA.SEN3",
            "Oa21_radiance.nc",
        )
        data = {
            "mission_id": "S3A",
            "datatype_id": "EFR",
            "start_time": dt.datetime(2018, 9, 16, 9, 5, 39),
            "end_time": dt.datetime(2018, 9, 16, 9, 8, 39),
            "creation_time": dt.datetime(2018, 9, 16, 9, 5, 39),
            "duration": 1,
            "cycle": 1,
            "relative_orbit": 1,
            "frame": 1,
            "centre": "CEN",
            "platform_mode": "M",
            "timeliness": "AA",
            "collection": "AAA",
            "dataset_name": "Oa21",
        }
        p = Parser(fmt)
        result = p.parse(filename)
        self.assertDictEqual(result, data)

    def test_parse_duplicate_fields(self):
        """Test parsing a pattern that has duplicate fields."""
        fmt = "{version_number:1s}/filename_with_version_number_{version_number:1s}.tif"
        filename = "1/filename_with_version_number_1.tif"
        p = Parser(fmt)
        result = p.parse(filename)
        self.assertEqual(result["version_number"], "1")



"""Basic unit tests for the parser module."""

import unittest
import datetime as dt
import pytest

from trollsift.parser import get_convert_dict, extract_values
from trollsift.parser import _convert
from trollsift.parser import parse, globify, validate, is_one2one, compose, Parser


class TestParser(unittest.TestCase):
    def setUp(self):
        self.fmt = "/somedir/{directory}/hrpt_{platform:4s}{platnum:2s}" + "_{time:%Y%m%d_%H%M}_{orbit:05d}.l1b"
        self.string = "/somedir/otherdir/hrpt_noaa16_20140210_1004_69022.l1b"
        self.string2 = "/somedir/otherdir/hrpt_noaa16_20140210_1004_00022.l1b"
        self.string3 = "/somedir/otherdir/hrpt_noaa16_20140210_1004_69022"
        self.string4 = "/somedir/otherdir/hrpt_noaa16_20140210_1004_69022"

    def test_parser_keys(self):
        parser = Parser(self.fmt)
        keys = {"directory", "platform", "platnum", "time", "orbit"}
        self.assertTrue(keys.issubset(parser.keys()) and keys.issuperset(parser.keys()))

    def test_get_convert_dict(self):
        # Run
        result = get_convert_dict(self.fmt)
        # Assert
        self.assertDictEqual(
            result,
            {
                "directory": "",
                "platform": "4s",
                "platnum": "2s",
                "time": "%Y%m%d_%H%M",
                "orbit": "05d",
            },
        )

    def test_extract_values(self):
        fmt = "/somedir/{directory}/hrpt_{platform:4s}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:d}.l1b"
        result = extract_values(fmt, self.string)
        self.assertDictEqual(
            result,
            {
                "directory": "otherdir",
                "platform": "noaa",
                "platnum": "16",
                "time": "20140210_1004",
                "orbit": "69022",
            },
        )

    def test_extract_values_end(self):
        fmt = "/somedir/{directory}/hrpt_{platform:4s}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:d}"
        result = extract_values(fmt, self.string3)
        self.assertDictEqual(
            result,
            {
                "directory": "otherdir",
                "platform": "noaa",
                "platnum": "16",
                "time": "20140210_1004",
                "orbit": "69022",
            },
        )

    def test_extract_values_beginning(self):
        fmt = "{directory}/hrpt_{platform:4s}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:d}"
        result = extract_values(fmt, self.string4)
        self.assertDictEqual(
            result,
            {
                "directory": "/somedir/otherdir",
                "platform": "noaa",
                "platnum": "16",
                "time": "20140210_1004",
                "orbit": "69022",
            },
        )

    def test_extract_values_s4spair(self):
        fmt = "{directory}/hrpt_{platform:4s}{platnum:s}_{time:%Y%m%d_%H%M}_{orbit:d}"
        result = extract_values(fmt, self.string4)
        self.assertDictEqual(
            result,
            {
                "directory": "/somedir/otherdir",
                "platform": "noaa",
                "platnum": "16",
                "time": "20140210_1004",
                "orbit": "69022",
            },
        )

    def test_extract_values_ss2pair(self):
        fmt = "{directory}/hrpt_{platform:s}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:d}"
        result = extract_values(fmt, self.string4)
        self.assertDictEqual(
            result,
            {
                "directory": "/somedir/otherdir",
                "platform": "noaa",
                "platnum": "16",
                "time": "20140210_1004",
                "orbit": "69022",
            },
        )

    def test_extract_values_ss2pair_end(self):
        fmt = "{directory}/hrpt_{platform:s}{platnum:2s}"
        result = extract_values(fmt, "/somedir/otherdir/hrpt_noaa16")
        self.assertDictEqual(
            result,
            {"directory": "/somedir/otherdir", "platform": "noaa", "platnum": "16"},
        )

    def test_extract_values_sdatetimepair_end(self):
        fmt = "{directory}/hrpt_{platform:s}{date:%Y%m%d}"
        result = extract_values(fmt, "/somedir/otherdir/hrpt_noaa20140212")
        self.assertDictEqual(
            result,
            {"directory": "/somedir/otherdir", "platform": "noaa", "date": "20140212"},
        )

    def test_extract_values_everything(self):
        fmt = "{everything}"
        result = extract_values(fmt, self.string)
        self.assertDictEqual(
            result,
            {"everything": "/somedir/otherdir/hrpt_noaa16_20140210_1004_69022.l1b"},
        )

    def test_extract_values_padding2(self):
        fmt = "/somedir/{directory}/hrpt_{platform:4s}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:0>5d}.l1b"
        # parsedef = ['/somedir/', {'directory': None}, '/hrpt_',
        #             {'platform': '4s'}, {'platnum': '2s'},
        #             '_', {'time': '%Y%m%d_%H%M'}, '_',
        #             {'orbit': '0>5d'}, '.l1b']
        result = extract_values(fmt, self.string2)
        # Assert
        self.assertDictEqual(
            result,
            {
                "directory": "otherdir",
                "platform": "noaa",
                "platnum": "16",
                "time": "20140210_1004",
                "orbit": "00022",
            },
        )

    def test_extract_values_fails(self):
        fmt = "/somedir/{directory}/hrpt_{platform:4s}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:4d}.l1b"
        self.assertRaises(ValueError, extract_values, fmt, self.string)

    def test_extract_values_full_match(self):
        """Test that a string must completely match."""
        fmt = "{orbit:05d}"
        val = extract_values(fmt, "12345")
        self.assertEqual(val, {"orbit": "12345"})
        self.assertRaises(ValueError, extract_values, fmt, "12345abc")
        val = extract_values(fmt, "12345abc", full_match=False)
        self.assertEqual(val, {"orbit": "12345"})

    def test_convert_digits(self):
        self.assertEqual(_convert("d", "69022"), 69022)
        self.assertRaises(ValueError, _convert, "d", "69dsf")
        self.assertEqual(_convert("d", "00022"), 22)
        self.assertEqual(_convert("4d", "69022"), 69022)
        self.assertEqual(_convert("_>10d", "_____69022"), 69022)
        self.assertEqual(_convert("%Y%m%d_%H%M", "20140210_1004"), dt.datetime(2014, 2, 10, 10, 4))

    def test_parse(self):
        # Run
        result = parse(self.fmt, "/somedir/avhrr/2014/hrpt_noaa19_20140212_1412_12345.l1b")
        # Assert
        self.assertDictEqual(
            result,
            {
                "directory": "avhrr/2014",
                "platform": "noaa",
                "platnum": "19",
                "time": dt.datetime(2014, 2, 12, 14, 12),
                "orbit": 12345,
            },
        )

    def test_parse_string_padding_syntax_with_and_without_s(self):
        """Test that, in string padding syntax, '' is equivalent to 's'.

        From <https://docs.python.org/3.4/library/string.html#format-specification-mini-language>:
            * Type 's': String format. This is the default type for strings and may be omitted.
            * Type None: The same as 's'.
        """
        result = parse("{foo}/{bar:_<8}", "baz/qux_____")
        expected_result = parse("{foo}/{bar:_<8s}", "baz/qux_____")
        self.assertEqual(expected_result["foo"], "baz")
        self.assertEqual(expected_result["bar"], "qux")
        self.assertEqual(result, expected_result)

    def test_parse_wildcards(self):
        # Run
        result = parse(
            "hrpt_{platform}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:05d}{ext}",
            "hrpt_noaa19_20140212_1412_12345.l1b",
        )
        # Assert
        self.assertDictEqual(
            result,
            {
                "platform": "noaa",
                "platnum": "19",
                "time": dt.datetime(2014, 2, 12, 14, 12),
                "orbit": 12345,
                "ext": ".l1b",
            },
        )

    def test_parse_align(self):
        filepattern = (
            "H-000-{hrit_format:4s}__-{platform_name:4s}________-"
            "{channel_name:_<9s}-{segment:_<9s}-{start_time:%Y%m%d%H%M}-__"
        )
        result = parse(filepattern, "H-000-MSG3__-MSG3________-IR_039___-000007___-201506051700-__")
        self.assertDictEqual(
            result,
            {
                "channel_name": "IR_039",
                "hrit_format": "MSG3",
                "platform_name": "MSG3",
                "segment": "000007",
                "start_time": dt.datetime(2015, 6, 5, 17, 0),
            },
        )

    def test_parse_digits(self):
        """Test when a digit field is shorter than the format spec."""
        result = parse(
            "hrpt_{platform}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:05d}{ext}",
            "hrpt_noaa19_20140212_1412_02345.l1b",
        )
        self.assertDictEqual(
            result,
            {
                "platform": "noaa",
                "platnum": "19",
                "time": dt.datetime(2014, 2, 12, 14, 12),
                "orbit": 2345,
                "ext": ".l1b",
            },
        )
        result = parse(
            "hrpt_{platform}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:5d}{ext}",
            "hrpt_noaa19_20140212_1412_ 2345.l1b",
        )
        self.assertDictEqual(
            result,
            {
                "platform": "noaa",
                "platnum": "19",
                "time": dt.datetime(2014, 2, 12, 14, 12),
                "orbit": 2345,
                "ext": ".l1b",
            },
        )
        result = parse(
            "hrpt_{platform}{platnum:2s}_{time:%Y%m%d_%H%M}_{orbit:_>5d}{ext}",
            "hrpt_noaa19_20140212_1412___345.l1b",
        )
        self.assertDictEqual(
            result,
            {
                "platform": "noaa",
                "platnum": "19",
                "time": dt.datetime(2014, 2, 12, 14, 12),
                "orbit": 345,
                "ext": ".l1b",
            },
        )

    def test_parse_bad_pattern(self):
        """Test when a digit field is shorter than the format spec."""
        self.assertRaises(
            ValueError,
            parse,
            "hrpt_{platform}{platnum:-=2s}_{time:%Y%m%d_%H%M}_{orbit:05d}{ext}",
            "hrpt_noaa19_20140212_1412_02345.l1b",
        )

    def test_globify_simple(self):
        # Run
        result = globify("{a}_{b}.end", {"a": "a", "b": "b"})
        # Assert
        self.assertEqual(result, "a_b.end")

    def test_globify_empty(self):
        # Run
        result = globify("{a}_{b:4d}.end", {})
        # Assert
        self.assertEqual(result, "*_????.end")

    def test_globify_noarg(self):
        # Run
        result = globify("{a}_{b:4d}.end")
        # Assert
        self.assertEqual(result, "*_????.end")

    def test_globify_known_lengths(self):
        # Run
        result = globify(
            "{directory}/{platform:4s}{satnum:2d}/{orbit:05d}",
            {"directory": "otherdir", "platform": "noaa"},
        )
        # Assert
        self.assertEqual(result, "otherdir/noaa??/?????")

    def test_globify_unknown_lengths(self):
        # Run
        result = globify(
            "hrpt_{platform_and_num}_" + "{date}_{time}_{orbit}.l1b",
            {"platform_and_num": "noaa16"},
        )
        # Assert
        self.assertEqual(result, "hrpt_noaa16_*_*_*.l1b")

    def test_globify_datetime(self):
        # Run
        result = globify(
            "hrpt_{platform}{satnum}_" + "{time:%Y%m%d_%H%M}_{orbit}.l1b",
            {"platform": "noaa", "time": dt.datetime(2014, 2, 10, 12, 12)},
        )
        # Assert
        self.assertEqual(result, "hrpt_noaa*_20140210_1212_*.l1b")

    def test_globify_partial_datetime(self):
        # Run
        result = globify(
            "hrpt_{platform:4s}{satnum:2d}_" + "{time:%Y%m%d_%H%M}_{orbit}.l1b",
            {"platform": "noaa", "time": (dt.datetime(2014, 2, 10, 12, 12), "Ymd")},
        )
        # Assert
        self.assertEqual(result, "hrpt_noaa??_20140210_????_*.l1b")

    def test_globify_datetime_nosub(self):
        # Run
        result = globify(
            "hrpt_{platform:4s}{satnum:2d}_" + "{time:%Y%m%d_%H%M}_{orbit}.l1b",
            {"platform": "noaa"},
        )

        # Assert
        self.assertEqual(result, "hrpt_noaa??_????????_????_*.l1b")

    def test_validate(self):
        # These cases are True
        self.assertTrue(validate(self.fmt, "/somedir/avhrr/2014/hrpt_noaa19_20140212_1412_12345.l1b"))
        self.assertTrue(validate(self.fmt, "/somedir/avhrr/2014/hrpt_noaa01_19790530_0705_00000.l1b"))
        self.assertTrue(validate(self.fmt, "/somedir/funny-char$dir/hrpt_noaa19_20140212_1412_12345.l1b"))
        self.assertTrue(validate(self.fmt, "/somedir//hrpt_noaa19_20140212_1412_12345.l1b"))
        # These cases are False
        self.assertFalse(validate(self.fmt, "/somedir/bla/bla/hrpt_noaa19_20140212_1412_1A345.l1b"))
        self.assertFalse(validate(self.fmt, "/somedir/bla/bla/hrpt_noaa19_2014021_1412_00000.l1b"))
        self.assertFalse(validate(self.fmt, "/somedir/bla/bla/hrpt_noaa19_20140212__412_00000.l1b"))
        self.assertFalse(validate(self.fmt, "/somedir/bla/bla/hrpt_noaa19_20140212__1412_00000.l1b"))
        self.assertFalse(validate(self.fmt, "/somedir/bla/bla/hrpt_noaa19_20140212_1412_00000.l1"))
        self.assertFalse(validate(self.fmt, "/somedir/bla/bla/hrpt_noaa19_20140212_1412_00000"))
        self.assertFalse(validate(self.fmt, "{}/somedir/bla/bla/hrpt_noaa19_20140212_1412_00000.l1b"))

    def test_is_one2one(self):
        # These cases are True
        self.assertTrue(is_one2one("/somedir/{directory}/somedata_{platform:4s}_{time:%Y%d%m-%H%M}_{orbit:5d}.l1b"))
        # These cases are False
        self.assertFalse(is_one2one("/somedir/{directory}/somedata_{platform:4s}_{time:%Y%d%m-%H%M}_{orbit:d}.l1b"))

    def test_greediness(self):
        """Test that the minimum match is parsed out.

        See GH #18.
        """
        from trollsift import parse

        template = "{band_type}_{polarization_extracted}_{unit}_{s1_fname}"
        fname = "Amplitude_VH_db_S1A_IW_GRDH_1SDV_20160528T171628_20160528T171653_011462_011752_0EED.tif"
        res_dict = parse(template, fname)
        exp = {
            "band_type": "Amplitude",
            "polarization_extracted": "VH",
            "unit": "db",
            "s1_fname": "S1A_IW_GRDH_1SDV_20160528T171628_20160528T171653_011462_011752_0EED.tif",
        }
        self.assertEqual(exp, res_dict)

        template = "{band_type:s}_{polarization_extracted}_{unit}_{s1_fname}"
        res_dict = parse(template, fname)
        self.assertEqual(exp, res_dict)


class TestCompose:
    """Test routines related to `compose` methods."""

    @pytest.mark.parametrize("allow_partial", [False, True])
    def test_compose(self, allow_partial):
        """Test the compose method's custom conversion options."""
        key_vals = {"a": "this Is A-Test b_test c test"}

        new_str = compose("{a!c}", key_vals, allow_partial=allow_partial)
        assert new_str == "This is a-test b_test c test"
        new_str = compose("{a!h}", key_vals, allow_partial=allow_partial)
        assert new_str == "thisisatestbtestctest"
        new_str = compose("{a!H}", key_vals, allow_partial=allow_partial)
        assert new_str == "THISISATESTBTESTCTEST"
        new_str = compose("{a!l}", key_vals, allow_partial=allow_partial)
        assert new_str == "this is a-test b_test c test"
        new_str = compose("{a!R}", key_vals, allow_partial=allow_partial)
        assert new_str == "thisIsATestbtestctest"
        new_str = compose("{a!t}", key_vals, allow_partial=allow_partial)
        assert new_str == "This Is A-Test B_Test C Test"
        new_str = compose("{a!u}", key_vals, allow_partial=allow_partial)
        assert new_str == "THIS IS A-TEST B_TEST C TEST"
        # builtin repr
        new_str = compose("{a!r}", key_vals, allow_partial=allow_partial)
        assert new_str == "'this Is A-Test b_test c test'"
        # no formatting
        new_str = compose("{a}", key_vals, allow_partial=allow_partial)
        assert new_str == "this Is A-Test b_test c test"
        # bad formatter
        with pytest.raises(ValueError):
            new_str = compose("{a!X}", key_vals, allow_partial=allow_partial)
        assert new_str == "this Is A-Test b_test c test"

    def test_default_compose_is_strict(self):
        """Make sure the default compose call does not accept partial composition."""
        fmt = "{foo}_{bar}.qux"
        with pytest.raises(KeyError):
            _ = compose(fmt, {"foo": "foo"})

    def test_partial_compose_simple(self):
        """Test partial compose with a simple use case."""
        fmt = "{variant:s}/{platform_name}_{start_time:%Y%m%d_%H%M}_{product}.{format}"
        composed = compose(
            fmt=fmt,
            keyvals={"platform_name": "foo", "format": "bar"},
            allow_partial=True,
        )
        assert composed == "{variant:s}/foo_{start_time:%Y%m%d_%H%M}_{product}.bar"

    def test_partial_compose_with_similarly_named_params(self):
        """Test that partial compose handles well vars with common substrings in name."""
        original_fmt = "{foo}{afooo}{fooo}.{bar}/{baz:%Y}/{baz:%Y%m%d_%H}/{baz:%Y}/{bar:d}"
        composed = compose(fmt=original_fmt, keyvals={"afooo": "qux"}, allow_partial=True)
        assert composed == "{foo}qux{fooo}.{bar}/{baz:%Y}/{baz:%Y%m%d_%H}/{baz:%Y}/{bar:d}"

    def test_partial_compose_repeated_vars_with_different_formatting(self):
        """Test partial compose with a fmt with repeated vars with different formatting."""
        fmt = "/foo/{start_time:%Y%m}/bar/{baz}_{start_time:%Y%m%d_%H%M}.{format}"
        composed = compose(fmt=fmt, keyvals={"format": "qux"}, allow_partial=True)
        assert composed == "/foo/{start_time:%Y%m}/bar/{baz}_{start_time:%Y%m%d_%H%M}.qux"

    @pytest.mark.parametrize(
        "original_fmt",
        ["{}_{}", "{foo}{afooo}{fooo}.{bar}/{baz:%Y}/{baz:%Y%m%d_%H}/{baz:%Y}/{bar:d}"],
    )
    def test_partial_compose_is_identity_with_empty_keyvals(self, original_fmt):
        """Test that partial compose leaves the input untouched if no keyvals at all."""
        assert compose(fmt=original_fmt, keyvals={}, allow_partial=True) == original_fmt

    def test_that_some_invalid_fmt_can_confuse_partial_compose(self):
        """Test that a fmt with a weird char can confuse partial compose."""
        fmt = "{foo?}_{bar}_{foo}.qux"
        with pytest.raises(ValueError):
            _ = compose(fmt=fmt, keyvals={}, allow_partial=True)


class TestParserFixedPoint:
    """Test parsing of fixed point numbers."""

    @pytest.mark.parametrize("allow_partial_compose", [False, True])
    @pytest.mark.parametrize(
        ("fmt", "string", "expected"),
        [
            # Naive
            ("{foo:f}", "12.34", 12.34),
            # Including width and precision
            ("{foo:5.2f}", "12.34", 12.34),
            ("{foo:5.2f}", "-1.23", -1.23),
            ("{foo:5.2f}", "12.34", 12.34),
            ("{foo:5.2f}", "123.45", 123.45),
            # Whitespace padded
            ("{foo:5.2f}", " 1.23", 1.23),
            ("{foo:5.2f}", " 12.34", 12.34),
            # Zero padded
            ("{foo:05.2f}", "01.23", 1.23),
            ("{foo:05.2f}", "012.34", 12.34),
            # Only precision, no width
            ("{foo:.2f}", "12.34", 12.34),
            # Only width, no precision
            ("{foo:16f}", "            1.12", 1.12),
            # No digits before decimal point
            ("{foo:3.2f}", ".12", 0.12),
            ("{foo:4.2f}", "-.12", -0.12),
            ("{foo:4.2f}", " .12", 0.12),
            ("{foo:4.2f}", "  .12", 0.12),
            ("{foo:16f}", "             .12", 0.12),
            # Exponential format
            ("{foo:7.2e}", "-1.23e4", -1.23e4),
        ],
    )
    def test_match(self, allow_partial_compose, fmt, string, expected):
        """Test cases expected to be matched."""
        # Test parsed value
        parsed = parse(fmt, string)
        assert parsed["foo"] == expected

        # Test round trip
        composed = compose(fmt, {"foo": expected}, allow_partial=allow_partial_compose)
        parsed = parse(fmt, composed)
        assert parsed["foo"] == expected

    @pytest.mark.parametrize(
        ("fmt", "string"),
        [
            # Decimals incorrect
            ("{foo:5.2f}", "12345"),
            ("{foo:5.2f}", "1234."),
            ("{foo:5.2f}", "1.234"),
            ("{foo:5.2f}", "123.4"),
            ("{foo:.2f}", "12.345"),
            # Decimals correct, but width too short
            ("{foo:5.2f}", "1.23"),
            ("{foo:5.2f}", ".23"),
            ("{foo:10.2e}", "1.23e4"),
            # Invalid
            ("{foo:5.2f}", "12_34"),
            ("{foo:5.2f}", "aBcD"),
        ],
    )
    def test_no_match(self, fmt, string):
        """Test cases expected to not be matched."""
        with pytest.raises(ValueError):
            parse(fmt, string)


@pytest.mark.parametrize(
    ("fmt", "string", "expected"),
    [
        # Decimal
        ("{foo:d}", "123", 123),
        # Hex with small letter
        ("{foo:x}", "7b", 123),
        # Hex with big letter
        ("{foo:X}", "7B", 123),
        # Fixed length hex
        ("{foo:03x}", "07b", 123),
        ("{foo:3x}", " 7b", 123),
        ("{foo:3X}", " 7B", 123),
        # Octal
        ("{foo:o}", "173", 123),
        # Free size with octal
        ("{bar:s}{foo:o}", "something173", 123),
        # Fixed length with octal
        ("{foo:_>4o}", "_173", 123),
        ("{foo:4o}", " 173", 123),
        # Binary
        ("{foo:b}", "1111011", 123),
        # Fixed length with binary
        ("{foo:8b}", " 1111011", 123),
        ("{foo:_>8b}", "_1111011", 123),
    ],
)
def test_parse_integers(fmt, string, expected):
    assert parse(fmt, string)["foo"] == expected



"""Main parsing and formatting functionality."""

from __future__ import annotations

import re
import datetime as dt
import random
import string
from functools import lru_cache
import typing

if typing.TYPE_CHECKING:
    from _typeshed import StrOrLiteralStr
    from typing import Any
    from collections.abc import Iterable, Sequence, Mapping


class Parser:
    """Class-based interface to parsing and formatting functionality."""

    def __init__(self, fmt: str):
        self.fmt = fmt

    def __str__(self):
        return self.fmt

    def keys(self):
        """Get parameter names defined in the format string."""
        convert_dict = get_convert_dict(self.fmt)
        return convert_dict.keys()

    def parse(self, stri: str, full_match: bool = True) -> dict[str, Any]:
        """Parse keys and values from ``stri`` using parser's format."""
        return parse(self.fmt, stri, full_match=full_match)

    def compose(self, keyvals: Mapping[str, Any], allow_partial: bool = False) -> str:
        """Compose format string ``self.fmt`` with parameters given in the ``keyvals`` dict.

        Args:
            keyvals: "Parameter --> parameter value" map
            allow_partial: If True, then partial composition is allowed, i.e.,
                not all parameters present in `fmt` need to be specified in `keyvals`.
                Unspecified parameters will, in this case, be left unchanged.
                (Default value = False).

        Returns:
            Result of formatting the *self.fmt* string with parameter values
            extracted from the corresponding items in the *keyvals* dictionary.

        """
        return compose(fmt=self.fmt, keyvals=keyvals, allow_partial=allow_partial)

    format = compose

    def globify(self, keyvals: Mapping[str, Any] | None = None) -> str:
        """Generate a string usable with glob.glob() from format string."""
        return globify(self.fmt, keyvals)

    def validate(self, stri: str) -> bool:
        """Validate that string ``stri`` conforms to the parser's format definition.

        Checks that the provided string is parsable and therefore complies with
        this parser's string format definition.  Useful for filtering strings,
        or to check if a string is compatible before passing it to the
        parser function.
        """
        return validate(self.fmt, stri)

    def is_one2one(self):
        """Check if this parser's format string has a one to one correspondence.

        That is, that successive composing and
        parsing operations will result in the original data.
        In other words, that input data maps to a string,
        which then maps back to the original data without any change
        or loss in information.

        Note: This test only applies to sensible usage of the format string.
        If string or numeric data causes overflow, e.g.
        if composing "abcd" into ``{3s}``, one to one correspondence will always
        be broken in such cases. This of course also applies to precision
        losses when using datetime data.
        """
        return is_one2one(self.fmt)


class StringFormatter(string.Formatter):
    """Custom string formatter class for basic strings.

    This formatter adds a few special conversions for assisting with common
    trollsift situations like making a parameter lowercase or removing
    hyphens. The added conversions are listed below and can be used in a
    format string by prefixing them with an `!` like so:

    >>> fstr = "{!u}_{!l}"
    >>> formatter = StringFormatter()
    >>> formatter.format(fstr, "to_upper", "To_LowerCase")
    "TO_UPPER_to_lowercase"

    - c: Make capitalized version of string (first character upper case, all lowercase after that) by executing the
      parameter's `.capitalize()` method.
    - l: Make all characters lowercase by executing the parameter's `.lower()` method.
    - R: Remove all separators from the parameter including '-', '_', ' ', and ':'.
    - t: Title case the string by executing the parameter's `.title()` method.
    - u: Make all characters uppercase by executing the parameter's `.upper()` method.
    - h: A combination of 'R' and 'l'.
    - H: A combination of 'R' and 'u'.

    """

    CONV_FUNCS = {
        "c": "capitalize",
        "h": "lower",
        "H": "upper",
        "l": "lower",
        "t": "title",
        "u": "upper",
    }

    def convert_field(self, value: str, conversion: str | None) -> str:
        """Apply conversions mentioned in `StringFormatter.CONV_FUNCS`."""
        if conversion is None:
            func = None
        else:
            func = self.CONV_FUNCS.get(conversion)
        if func is not None:
            value = getattr(value, func)()
        elif conversion not in ["R"]:
            # default conversion ('r', 's')
            return super(StringFormatter, self).convert_field(value, conversion)

        if conversion in ["h", "H", "R"]:
            value = value.replace("-", "").replace("_", "").replace(":", "").replace(" ", "")
        return value


formatter = StringFormatter()


# taken from https://docs.python.org/3/library/re.html#simulating-scanf
spec_regexes = {
    "b": r"[-+]?[0-1]",
    "c": r".",
    "d": r"[-+]?\d",
    # Naive fixed point format specifier (e.g. {foo:f})
    "f": r"[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?",
    # Fixed point format specifier including width and precision
    # (e.g. {foo:4.2f}). The lookahead (?=.{width}) makes sure that the
    # subsequent pattern is only matched if the string has the required
    # (minimum) width.
    "f_with_precision": r"(?=.{{{width}}})([-+]?([\d ]+(\.\d{{{decimals}}})+|\.\d{{{decimals}}})([eE][-+]?\d+)?)",
    "i": r"[-+]?(0[xX][\dA-Fa-f]+|0[0-7]*|\d+)",
    "o": r"[-+]?[0-7]",
    "s": r"\S",
    "x": r"[-+]?(0[xX])?[\dA-Fa-f]",
}
spec_regexes["e"] = spec_regexes["f"]
spec_regexes["E"] = spec_regexes["f"]
spec_regexes["g"] = spec_regexes["f"]
spec_regexes["X"] = spec_regexes["x"]
spec_regexes[""] = spec_regexes["s"]
allow_multiple = ["b", "c", "d", "o", "s", "", "x", "X"]
fixed_point_types = ["f", "e", "E", "g"]
# format_spec ::=  [[fill]align][sign][#][0][width][,][.precision][type]
# https://docs.python.org/3.4/library/string.html#format-specification-mini-language
fmt_spec_regex = re.compile(
    r"(?P<align>(?P<fill>.)?[<>=^])?(?P<sign>[\+\-\s])?(?P<pound>#)?(?P<zero>0)?(?P<width>\d+)?"
    r"(?P<comma>,)?(?P<precision>.\d+)?(?P<type>[bcdeEfFgGnosxX%]?)"
)


def _get_fixed_point_regex(width: str | None, precision: str | None) -> str:
    """Get regular expression for fixed point numbers.

    Args:
        width: Total width of the string representation.
        precision: Number of decimals.
    """
    if width or precision:
        if precision is None:
            precision = "0,"
        else:
            precision = precision.strip(".")
        if width is None:
            width = "1,"
        return spec_regexes["f_with_precision"].format(width=width, decimals=precision)
    else:
        return spec_regexes["f"]


class RegexFormatter(string.Formatter):
    """String formatter that converts a format string to a regular expression.

    >>> regex_formatter = RegexFormatter()
    >>> regex_str = regex_formatter.format('{field_one:5d}_{field_two}')

    Can also be used to extract values from a string given the format spec
    for that string:

    >>> regex_formatter.extract_values('{field_one:5d}_{field_two}', '12345_sometext')
    {'field_one': '12345', 'field_two': 'sometext'}

    Note that the regular expressions generated by this class are specially
    generated to reduce "greediness" of the matches found. For ambiguous
    patterns where a single field could match shorter or longer portions of
    the provided string, this class will prefer the shorter version of the
    string in order to make the rest of the pattern match. For example:

    >>> regex_formatter.extract_values('{field_one}_{field_two}', 'abc_def_ghi')
    {'field_one': 'abc', 'field_two': 'def_ghi'}

    Note how `field_one` could have matched "abc_def", but the lower
    greediness of this parser caused it to only match against "abc".

    """

    # special string to mark a parameter not being specified
    UNPROVIDED_VALUE = "<trollsift unprovided value>"
    ESCAPE_CHARACTERS = ["\\"] + [x for x in string.punctuation if x not in "\\%"]
    ESCAPE_SETS = [(c, "\\" + c) for c in ESCAPE_CHARACTERS]

    def __init__(self):
        # hold on to fields we've seen already so we can reuse their
        # definitions in the regex
        self._cached_fields = {}
        self.format = lru_cache()(self._uncached_format)
        super(RegexFormatter, self).__init__()

    def _uncached_format(*args, **kwargs):
        try:
            # super() doesn't seem to work here
            ret_val = string.Formatter.format(*args, **kwargs)
        finally:
            self = args[0]  # just matching the parent class
            self._cached_fields.clear()
        return ret_val

    def _escape(self, s: str) -> str:
        """Escape bad characters for regular expressions.

        Similar to `re.escape` but allows '%' to pass through.

        """
        for ch, r_ch in self.ESCAPE_SETS:
            s = s.replace(ch, r_ch)
        return s

    def parse(
        self, format_string: StrOrLiteralStr
    ) -> Iterable[
        tuple[
            StrOrLiteralStr,
            StrOrLiteralStr | None,
            StrOrLiteralStr | None,
            StrOrLiteralStr | None,
        ]
    ]:
        parse_ret = super(RegexFormatter, self).parse(format_string)
        for literal_text, field_name, format_spec, conversion in parse_ret:
            # the parent class will call parse multiple times moving
            # 'format_spec' to 'literal_text'. We only escape 'literal_text'
            # so we don't escape things twice.
            literal_text = self._escape(literal_text)
            yield literal_text, field_name, format_spec, conversion

    def get_value(self, key: int | str, args: Sequence[Any], kwargs: Mapping[str, Any]) -> Any:
        try:
            return super(RegexFormatter, self).get_value(key, args, kwargs)
        except (IndexError, KeyError):
            return key, self.UNPROVIDED_VALUE

    def _regex_datetime(self, format_spec: str) -> str:
        replace_str = format_spec
        for fmt_key, fmt_val in DT_FMT.items():
            if fmt_key == "%%":
                # special case
                replace_str.replace("%%", "%")
                continue
            count = fmt_val.count("?")
            # either a series of numbers or letters/numbers
            regex = r"\d{{{:d}}}".format(count) if count else r"[^ \t\n\r\f\v\-_:]+"
            replace_str = replace_str.replace(fmt_key, regex)
        return replace_str

    def regex_field(self, field_name: str, value: Any, format_spec: str) -> str:
        if value != self.UNPROVIDED_VALUE:
            return super(RegexFormatter, self).format_field(value, format_spec)

        if self._cached_fields.get(field_name, format_spec) != format_spec:
            raise ValueError("Can't specify the same field_name with different formats: {}".format(field_name))
        elif field_name in self._cached_fields:
            return r"(?P={})".format(field_name)
        else:
            self._cached_fields[field_name] = format_spec

        # Replace format spec with glob patterns (*, ?, etc)
        if not format_spec:
            return r"(?P<{}>.*?)".format(field_name)
        if "%" in format_spec:
            return r"(?P<{}>{})".format(field_name, self._regex_datetime(format_spec))
        return format_spec_to_regex(field_name, format_spec)

    def format_field(self, value: Any, format_spec: str) -> str:
        if not isinstance(value, tuple) or value[1] != self.UNPROVIDED_VALUE:
            return super(RegexFormatter, self).format_field(value, format_spec)
        field_name, value = value
        return self.regex_field(field_name, value, format_spec)


def format_spec_to_regex(field_name: str, format_spec: str) -> str:
    """Make an attempt at converting a format spec to a regular expression."""
    # NOTE: remove escaped backslashes so regex matches
    regex_match = fmt_spec_regex.match(format_spec.replace("\\", ""))
    if regex_match is None:
        raise ValueError("Invalid format specification: '{}'".format(format_spec))
    regex_dict = regex_match.groupdict()
    ftype = regex_dict["type"]
    width = regex_dict["width"]
    align = regex_dict["align"]
    precision = regex_dict["precision"]
    fill = _get_fill(regex_dict["fill"], width, ftype)

    char_type = spec_regexes[ftype]
    if ftype in fixed_point_types:
        char_type = _get_fixed_point_regex(width=width, precision=precision)
    if ftype in ("s", "") and align and align.endswith("="):
        raise ValueError("Invalid format specification: '{}'".format(format_spec))
    final_regex = char_type
    if ftype in allow_multiple and (not width or width == "0"):
        final_regex += r"*?"
    elif width and width != "0":
        if not fill and ftype not in fixed_point_types:
            # we know we have exactly this many characters
            final_regex += r"{{{}}}".format(int(width))
        elif fill:
            # we don't know how many fill characters we have compared to
            # field characters so just match all characters and sort it out
            # later during type conversion.
            final_regex = r".{{{}}}".format(int(width))
        elif ftype in allow_multiple:
            final_regex += r"*?"

    return r"(?P<{}>{})".format(field_name, final_regex)


def _get_fill(fill: str | None, width: str | None, ftype: str | None) -> str | None:
    # NOTE: does not properly handle `=` alignment
    if fill is None:
        if width is not None and width[0] == "0":
            fill = "0"
        elif ftype in ["s", "", "d", "x", "X", "o", "b"]:
            fill = " "
    return fill


@lru_cache()
def regex_format(fmt: str) -> str:
    # We create a new instance of RegexFormatter here to prevent concurrent calls to
    # format interfering with one another.
    return RegexFormatter().format(fmt)


def extract_values(fmt: str, stri: str, full_match: bool = True) -> dict[str, Any]:
    """Extract information from string matching format.

    Args:
        fmt: Python format string to match against
        stri: String to extract information from
        full_match: Force the match of the whole string. Default
            to ``True``.
    """
    regex = regex_format(fmt)
    if full_match:
        regex = "^" + regex + "$"
    match = re.match(regex, stri)
    if match is None:
        raise ValueError("String does not match pattern.")
    return match.groupdict()


def _get_number_from_fmt(fmt: str) -> int:
    """Helper function for extract_values.

    Figures out string length from format string.

    """
    if "%" in fmt:
        # its datetime
        return len(("{0:" + fmt + "}").format(dt.datetime.now()))
    else:
        # its something else
        fmt = fmt.lstrip("0")
        fmt_digits_match = re.search("[0-9]+", fmt)
        if fmt_digits_match is None:
            raise ValueError(f"No number specified in format string: {fmt}")
        return int(fmt_digits_match.group(0))


def _convert(convdef: str, stri: str) -> Any:
    """Convert the string *stri* to the given conversion definition *convdef*."""
    result: Any  # force mypy type
    if "%" in convdef:
        result = dt.datetime.strptime(stri, convdef)
    else:
        result = _strip_padding(convdef, stri)
        if "d" in convdef:
            result = int(result)
        elif "x" in convdef or "X" in convdef:
            result = int(result, 16)
        elif "o" in convdef:
            result = int(result, 8)
        elif "b" in convdef:
            result = int(result, 2)
        elif any(float_type_marker in convdef for float_type_marker in fixed_point_types):
            result = float(result)

    return result


def _strip_padding(convdef: str, stri: str) -> str:
    """Strip padding from the given string.

    Args:
        convdef: Conversion definition (indicates the padding)
        stri: String to be modified
    """
    regex_match = fmt_spec_regex.match(convdef)
    match_dict = regex_match.groupdict() if regex_match else {}
    align = match_dict.get("align")
    pad = match_dict.get("fill")
    if align:
        # align character is the last one
        align = align[-1]
    if align and align in "<>^" and not pad:
        pad = " "
    if align == ">":
        stri = stri.lstrip(pad)
    elif align == "<":
        stri = stri.rstrip(pad)
    elif align == "^":
        stri = stri.strip(pad)
    return stri


@lru_cache()
def get_convert_dict(fmt: str) -> dict[str, str]:
    """Retrieve parse definition from the format string `fmt`."""
    convdef = {}
    for _literal_text, field_name, format_spec, _conversion in formatter.parse(fmt):
        if field_name is None or format_spec is None:
            continue
        # XXX: Do I need to include 'conversion'?
        convdef[field_name] = format_spec
    return convdef


def parse(fmt: str, stri: str, full_match: bool = True) -> dict[str, Any]:
    """Parse keys and corresponding values from *stri* using format described in *fmt* string.

    Args:
        fmt: Python format string to match against
        stri: String to extract information from
        full_match: Force the match of the whole string. Default True.

    """
    convdef = get_convert_dict(fmt)
    keyvals = extract_values(fmt, stri, full_match=full_match)
    for key in convdef.keys():
        keyvals[key] = _convert(convdef[key], keyvals[key])

    return keyvals


def compose(fmt: str, keyvals: Mapping[str, Any], allow_partial: bool = False) -> str:
    """Compose format string *self.fmt* with parameters given in the *keyvals* dict.

    Args:
        fmt: Python format string to match against
        keyvals: "Parameter --> parameter value" map
        allow_partial: If True, then partial composition is allowed, i.e.,
            not all parameters present in `fmt` need to be specified in `keyvals`.
            Unspecified parameters will, in this case, be left unchanged.
            (Default value = False).

    Returns:
        Result of formatting the *self.fmt* string with parameter values
        extracted from the corresponding items in the *keyvals* dictionary.

    """
    if allow_partial:
        return _partial_compose(fmt=fmt, keyvals=keyvals)
    return _strict_compose(fmt=fmt, keyvals=keyvals)


DT_FMT = {
    "%a": "*",
    "%A": "*",
    "%w": "?",
    "%d": "??",
    "%b": "*",
    "%B": "*",
    "%m": "??",
    "%y": "??",
    "%Y": "????",
    "%H": "??",
    "%I": "??",
    "%p": "*",
    "%M": "??",
    "%S": "??",
    "%f": "*",
    "%z": "*",
    "%Z": "*",
    "%j": "???",
    "%U": "??",
    "%W": "??",
    "%c": "*",
    "%x": "*",
    "%X": "*",
    "%%": "?",
}


class GlobifyFormatter(string.Formatter):
    # special string to mark a parameter not being specified
    UNPROVIDED_VALUE = "<trollsift unprovided value>"

    def get_value(self, key: str | int, args: Sequence[Any], kwargs: Mapping[str, Any]) -> Any:
        try:
            return super(GlobifyFormatter, self).get_value(key, args, kwargs)
        except (IndexError, KeyError):
            # assumes that
            return self.UNPROVIDED_VALUE

    def format_field(self, value: Any, format_spec: str) -> str:
        if not isinstance(value, (list, tuple)) and value != self.UNPROVIDED_VALUE:
            return super(GlobifyFormatter, self).format_field(value, format_spec)
        elif value != self.UNPROVIDED_VALUE:
            # partial provided date/time fields
            # specified with a tuple/list of 2 elements
            # (value, partial format string)
            value, dt_fmt = value
            for fmt_letter in dt_fmt:
                fmt = "%" + fmt_letter
                format_spec = format_spec.replace(fmt, value.strftime(fmt))

        # Replace format spec with glob patterns (*, ?, etc)
        if not format_spec:
            return "*"
        if "%" in format_spec:
            replace_str = format_spec
            for fmt_key, fmt_val in DT_FMT.items():
                replace_str = replace_str.replace(fmt_key, fmt_val)
            return replace_str
        if not re.search("[0-9]+", format_spec):
            # non-integer type
            return "*"
        return "?" * _get_number_from_fmt(format_spec)


globify_formatter = GlobifyFormatter()


def globify(fmt: str, keyvals: Mapping[str, Any] | None = None) -> Any:
    """Generate a string usable with glob.glob() from format string and provided information."""
    if keyvals is None:
        keyvals = {}
    return globify_formatter.format(fmt, **keyvals)


def validate(fmt: str, stri: str) -> bool:
    """Validates that string ``stri`` conforms to ``fmt``.

    Useful for filtering string, or to check if string is compatible before
    passing the string to the parser function.
    """
    try:
        parse(fmt, stri)
        return True
    except ValueError:
        return False


def _generate_data_for_format(fmt: str) -> dict[str, Any]:
    """Generate a fake data dictionary to fill in the provided format string."""
    # finally try some data, create some random data for the fmt.
    data = {}
    # keep track of how many "free_size" (wildcard) parameters we have
    # if we get two in a row then we know the pattern is invalid, meaning
    # we'll never be able to match the second wildcard field
    free_size_start = False
    for literal_text, field_name, format_spec, _conversion in formatter.parse(fmt):
        if literal_text:
            free_size_start = False

        if not field_name:
            free_size_start = False
            continue

        # encapsulating free size keys,
        # e.g. {:s}{:s} or {:s}{:4s}{:d}
        if not format_spec or format_spec == "s" or format_spec == "d":
            if free_size_start:
                raise ValueError("Can't generate data for spec with two or more fields with no size specifier.")
            else:
                free_size_start = True

        # make some data for this key and format
        data[field_name] = _gen_data_for_spec(format_spec)
    return data


def _gen_data_for_spec(format_spec: str | None) -> int | str | dt.datetime:
    if format_spec and "%" in format_spec:
        # some datetime
        t = dt.datetime.now()
        # run once through format to limit precision
        t = parse("{t:" + format_spec + "}", compose("{t:" + format_spec + "}", {"t": t}))["t"]
        return t

    if format_spec and "d" in format_spec:
        # random number (with n sign. figures)
        if not format_spec.isalpha():
            n = _get_number_from_fmt(format_spec)
        else:
            # clearly bad
            raise ValueError(f"Bad format specification: {format_spec!r}")
        return random.randint(0, 99999999999999999) % (10**n)

    # string type
    if format_spec is None:
        n = 4
    elif format_spec.isalnum():
        n = _get_number_from_fmt(format_spec)
    else:
        n = 4
    randstri = ""
    for _ in range(n):
        randstri += random.choice(string.ascii_letters)
    return randstri


def is_one2one(fmt: str) -> bool:
    """Check if the format string has a one to one correspondence.

    That is, that successive composing and
    parsing operations will result in the original data.
    In other words, that input data maps to a string,
    which then maps back to the original data without any change
    or loss in information.

    Note: This test only applies to sensible usage of the format string.
    If string or numeric data is causes overflow, e.g.
    if composing "abcd" into {3s}, one to one correspondence will always
    be broken in such cases. This of course also applies to precision
    losses when using  datetime data.
    """
    try:
        data = _generate_data_for_format(fmt)
    except ValueError:
        return False

    # run data forward once and back to data
    stri = compose(fmt, data)
    data2 = parse(fmt, stri)
    # check if data2 equal to original data
    if len(data) != len(data2):
        return False
    for key in data:
        if key not in data2:
            return False
        if data2[key] != data[key]:
            return False
    # all checks passed, so just return True
    return True


def purge() -> None:
    """Clear internal caches.

    Not needed normally, but can be used to force cache clear when memory
    is very limited.

    """
    regex_format.cache_clear()
    get_convert_dict.cache_clear()


def _strict_compose(fmt: str, keyvals: Mapping[str, Any]) -> str:
    """Convert parameters in `keyvals` to a string based on `fmt` string."""
    return formatter.format(fmt, **keyvals)


def _partial_compose(fmt: str, keyvals: Mapping[str, Any]) -> str:
    """Convert parameters in `keyvals` to a string based on `fmt` string.

    Similar to _strict_compose, but accepts partial composing, i.e., not all
    parameters in `fmt` need to be specified in `keyvals`. Unspecified parameters
    are left unchanged.

    Args:
        fmt (str): Python format string to match against
        keyvals (dict): "Parameter --> parameter value" map

    """
    fmt, undefined_vars = _replace_undefined_params_with_placeholders(fmt, keyvals)
    composed_string = _strict_compose(fmt=fmt, keyvals=keyvals)
    for fmt_placeholder, fmt_specification in undefined_vars.items():
        composed_string = composed_string.replace(fmt_placeholder, fmt_specification)

    return composed_string


def _replace_undefined_params_with_placeholders(
    fmt: str, keyvals: Mapping[str, Any] | None = None
) -> tuple[str, dict[str, Any]]:
    """Replace with placeholders params in `fmt` not specified in `keyvals`."""
    vars_left_undefined = set(get_convert_dict(fmt).keys())
    if keyvals is not None:
        vars_left_undefined -= keyvals.keys()

    undefined_vars_placeholders_dict = {}
    new_fmt = fmt
    for var in sorted(vars_left_undefined):
        matches = set(match.group() for match in re.finditer(rf"{{{re.escape(var)}([^\w{{}}].*?)*}}", new_fmt))
        if len(matches) == 0:
            raise ValueError(f"Could not capture definitions for {var} from {fmt}")
        for var_specification in matches:
            fmt_placeholder = f"({hex(hash(var_specification))})"
            undefined_vars_placeholders_dict[fmt_placeholder] = var_specification
            new_fmt = new_fmt.replace(var_specification, fmt_placeholder)

    return new_fmt, undefined_vars_placeholders_dict



"""Parser regression tests."""

import unittest
import datetime as dt

from trollsift.parser import parse


class TestParser(unittest.TestCase):
    def test_002(self):
        res = parse(
            "hrpt16_{satellite:7s}_{start_time:%d-%b-%Y_%H:%M:%S.000}_{orbit_number:5d}",
            "hrpt16_NOAA-19_26-NOV-2014_10:12:00.000_29889",
        )
        self.assertEqual(
            res,
            {
                "orbit_number": 29889,
                "satellite": "NOAA-19",
                "start_time": dt.datetime(2014, 11, 26, 10, 12),
            },
        )



from .parser import Parser, StringFormatter, parse, compose, globify, purge, validate

try:
    from trollsift.version import version as __version__  # noqa
except ModuleNotFoundError:  # pragma: no cover
    raise ModuleNotFoundError(
        "No module named trollsift.version. This could mean "
        "you didn't install 'trollsift' properly. Try reinstalling ('pip "
        "install')."
    ) from None

__all__ = [
    "Parser",
    "StringFormatter",
    "parse",
    "compose",
    "globify",
    "purge",
    "validate",
]
